
#ifndef BFM_LINALG_H
#define BFM_LINALG_H

/*  2007
 *  Calls PAB's assembler routines to give an implementation
 *  of the dwf dslash. Aim is to give very high performance
 *  in a reasonably portable/retargettable library.
 */
#include "bfm.h"
#include <stdio.h>
#include <stdlib.h>

static double aa[4] __attribute__ ((aligned(32)));
static double bb[4] __attribute__ ((aligned(32)));


template <class Float>
double bfmbase<Float>::axpby_norm(Fermion_t r, Fermion_t x, Fermion_t y,double a,double b)
{
  Float *xx = (Float * )x;
  Float *yy = (Float * )y;
  Float *rr = (Float * )r;
  double nrm=0.0;
  integer len = 4*cbLs*simd_cbvol*nsimd;

  if ( sizeof(Float) == sizeof(double) ) {
    vmx_vaxpby_norm((double *)rr,&a,&b,(double *)xx,(double *)yy,len,&nrm);
  } else { 
    vmx_vaxpby_norm_single((float *)rr,&a,&b,(float *)xx,(float *)yy,len,&nrm);
  }
  repro(nrm);
  comm_gsum(nrm);
  return nrm;
}
template <class Float>
double bfmbase<Float>::axpy_norm(Fermion_t r, Fermion_t x, Fermion_t y,double a)
{
  Float *xx = (Float * )x;
  Float *yy = (Float * )y;
  Float *rr = (Float * )r;
  double nrm=0.0;
  integer len = 4*cbLs*simd_cbvol*nsimd;
  if ( sizeof(Float) == sizeof(double) ) {
    vmx_vaxpy_norm((double *)rr,&a,(double *)xx,(double *)yy,len,&nrm);
  } else { 
    vmx_vaxpy_norm_single((float *)rr,&a,(float *)xx,(float *)yy,len,&nrm);
  }
  repro(nrm);
  comm_gsum(nrm);
  return nrm;
}
template <class Float>
void bfmbase<Float>::axpby(Fermion_t r, Fermion_t x, Fermion_t y,double a,double b)
{
  Float *xx = (Float * )x;
  Float *yy = (Float * )y;
  Float *rr = (Float * )r;
  integer len = 2*cbLs*simd_cbvol;
  aa[0] = aa[1] = aa[2] = aa[3] = a;
  bb[0] = bb[1] = bb[2] = bb[3] = b;
  if ( sizeof(Float) == sizeof(double) ) {
    vmx_vaxpby((double *)rr,(double *)aa,(double *)bb,(double *)xx,(double *)yy,len);
  } else { 
    vmx_vaxpby_single((float *)rr,(double *)aa,(double *)bb,(float *)xx,(float *)yy,len);
  }
}
template <class Float>
void bfmbase<Float>::axpy(Fermion_t r, Fermion_t x, Fermion_t y,double a)
{
  Float *xx = (Float * )x;
  Float *yy = (Float * )y;
  Float *rr = (Float * )r;
  integer len = 2*cbLs*simd_cbvol;
  aa[0] = aa[1] = aa[2] = aa[3] = a;
  if ( sizeof(Float) == sizeof(double) ) {
    vmx_vaxpy((double *)rr,(double *)aa,(double *)xx,(double *)yy,len);
  } else { 
    vmx_vaxpy_single((float *)rr,(double *)aa,(float *)xx,(float *)yy,len);
  } 
  return;
}

template <class Float>
void bfmbase<Float>::scale(Fermion_t x_t, double a)
{
  Float *x = (Float *) x_t;
  int nspinco = 12;
  Float *xx;
  Float nrm = 0.0;

  for(int site=0;site<simd_cbvol;site++){
   for(int s=0;s<Ls;s++) {

    xx = &x[s*nspinco*nsimd*2   + site*nspinco*nsimd*2*Ls];

    for(int spinco=0;spinco<nspinco;spinco++){
    for(int simd_el=0;simd_el<nsimd;simd_el++){
    for(int reim=0;reim<2;reim++){
      int idx = reim + simd_el*2 + spinco*nsimd*2;
      xx[idx] = a *xx[idx];
    }}}
   }
  }  
}
template <class Float>
double bfmbase<Float>::norm(Fermion_t x_t)
{
  Float *x = (Float *) x_t;
  int nspinco = 12;
  Float *xx;
  double nrm = 0.0;

  for(int site=0;site<simd_cbvol;site++){
   for(int s=0;s<cbLs;s++) {

    xx = &x[s*nspinco*nsimd*2   + site*nspinco*nsimd*2*cbLs];

    for(int spinco=0;spinco<nspinco;spinco++){
    for(int simd_el=0;simd_el<nsimd;simd_el++){
    for(int reim=0;reim<2;reim++){
      int idx = reim + simd_el*2 + spinco*nsimd*2;
      nrm = nrm+ xx[idx]*xx[idx];
    }}}
   }
  }  
  repro(nrm);
  comm_gsum(nrm);
  return nrm;
}
template <class Float>
void bfmbase<Float>::chiralproj_axpby(Fermion_t y, Fermion_t x,double a, int sign, int ls_slices)
{
  Float *xx;
  Float *yy;

  int nspinco = 12;
  integer len = ls_slices*simd_cbvol*nsimd;

  //  if ( len & 0x1 ) exit(-1);
  //  len = len/2;

  /*
   * 1+g5 -> 1 0
   *         0 0
   * Offset => 0 if sign == -1 
   * 
   * 1-g5 -> 0 0
   *         0 1
   * Offset => 12 if sign == -1 
   *   
   */      
  if ( sign == 1 ) { 
    xx = (Float * )x;
    yy = (Float * )y;
  } else if ( sign == -1 ) { 
    xx = ((Float * )x) + nspinco*nsimd;
    yy = ((Float * )y) + nspinco*nsimd;
  } else { 
    printf("oops - bad sign in chiralproj_axpby\n");
    exit(0);
  }
  
  aa[0] = aa[1] = aa[2] = aa[3] = a;
  
  if ( sizeof(Float) == sizeof(double) ) {
    vmx_1pG5vaxpy((double *)yy,(double *)aa,(double *)xx,len);
  } else { 
    vmx_1pG5vaxpy_single((float *)yy,(double *)aa,(float *)xx,len);
  }

}

#endif
