#ifndef BFM_QDP_H
#define BFM_QDP_H

#include <bfm.h>

#ifdef BFM_QDP_BINDING

template <class Float_f,class Float_d>
int dwf_CG(LatticePropagator &sol, 
	   LatticePropagator &src,
	   multi1d<LatticeColorMatrix> &U,
	   std::string output_stem,
	   int Ls,
	   int DodwfConservedVector,
	   Real mass,
	   Real M5,int Ncg, Real resid[],int max_iter[]);

template <class Float>
int dwf_invert(multi1d<LatticeFermion> &sol, 
	       multi1d<LatticeFermion> &src,
	       multi1d<LatticeColorMatrix> &U,
	       std::string output_stem,
	       int Ls,
	       Real mass,
	       Real M5, Real residual, int max_iter);

template <class Float_f,class Float_d>
int dwf_restarted_invert(multi1d<LatticeFermion> &sol, 
			 multi1d<LatticeFermion> &src,
			 multi1d<LatticeColorMatrix> &U,
			 std::string output_stem,
			 int Ls,
			 Real mass,
			 Real M5,
			 int Ncg,
			 Real residual[],
			 int max_iter[]);

#define BFM_CHROMA_BINDING
#ifdef BFM_CHROMA_BINDING
#include <chroma.h>
#else
/* FermToProp and PropToFerm lifted from CHROMA - sorry guys*/
/* You lifted my routines in what you call bagel-qdp anyway... touche*/
  void FermToProp(const LatticeFermionF& a, LatticePropagatorF& b, 
		  int color_index, int spin_index);
  //! Extract a LatticeFermion from a LatticePropagator
  /*!
   * \ingroup ferm
   *
   * \param a      Source propagator
   * \param b      Destination fermion
   * \param color_index  Color index
   * \param spin_index   Spin index
   */
  void PropToFerm(const LatticePropagatorF& b, LatticeFermionF& a, 
		  int color_index, int spin_index)
  {
    for(int j = 0; j < Ns; ++j)
    {
      LatticeColorMatrixF bb = peekSpin(b, j, spin_index);
      LatticeColorVectorF aa = peekSpin(a, j);

      for(int i = 0; i < Nc; ++i)
	pokeColor(aa, peekColor(bb, i, color_index), i);

      pokeSpin(a, aa, j);
    }
  }

  //! Extract a LatticeFermion from a LatticePropagator
  /*!
   * \ingroup ferm
   *
   * \param a      Source propagator
   * \param b      Destination fermion
   * \param color_index  Color index
   * \param spin_index   Spin index
   */
  void PropToFerm(const LatticePropagatorD& b, LatticeFermionD& a, 
		  int color_index, int spin_index)
  {
    for(int j = 0; j < Ns; ++j)
    {
      LatticeColorMatrixD bb = peekSpin(b, j, spin_index);
      LatticeColorVectorD aa = peekSpin(a, j);

      for(int i = 0; i < Nc; ++i)
	pokeColor(aa, peekColor(bb, i, color_index), i);

      pokeSpin(a, aa, j);
    }
  }


  void FermToProp(const LatticeFermionF& a, LatticePropagatorF& b, 
		  int color_index, int spin_index)
  {
    for(int j = 0; j < Ns; ++j)
    {
      LatticeColorMatrixF bb = peekSpin(b, j, spin_index);
      LatticeColorVectorF aa = peekSpin(a, j);

      for(int i = 0; i < Nc; ++i)
	pokeColor(bb, peekColor(aa, i), i, color_index);

      pokeSpin(b, bb, j, spin_index);
    }
  }

  //! Insert a LatticeFermion into a LatticePropagator
  /*!
   * \ingroup ferm
   *
   * \param a      Source fermion
   * \param b      Destination propagator
   * \param color_index  Color index
   * \param spin_index   Spin index
   */
  void FermToProp(const LatticeFermionD& a, LatticePropagatorD& b, 
		  int color_index, int spin_index)
  {
    for(int j = 0; j < Ns; ++j)
    {
      LatticeColorMatrixD bb = peekSpin(b, j, spin_index);
      LatticeColorVectorD aa = peekSpin(a, j);

      for(int i = 0; i < Nc; ++i)
	pokeColor(bb, peekColor(aa, i), i, color_index);

      pokeSpin(b, bb, j, spin_index);
    }
  }
#endif

#ifdef CHROMA_3
static Set timeslice;
#else
static UnorderedSet timeslice;
#endif
static int timeslice_is_made = 0;

class TSliceFunc : public SetFunc
{
public:
  TSliceFunc(int dir): dir_decay(dir) {}

  int operator() (const multi1d<int>& coordinate) const {return coordinate[dir_decay];}
  int numSubsets() const {return Layout::lattSize()[dir_decay];}

  int dir_decay;

private:
  TSliceFunc() {}  // hide default constructor
};
Set &GetTimeslice(void);


// Internal interfaces
template <class Float_f,class Float_d>
int dwf_CG(LatticePropagator &sol, 
	   LatticePropagator &src,
	   multi1d<LatticeColorMatrix> &U,
	   std::string output_stem,
	   int Ls,
	   int do_PImunu,
	   Real mass,
	   Real M5,int Ncg,
	   Real residual[],
	   int max_iter[])
{

  std::string output    = output_stem + ".xml";
  std::string outputbin = output_stem + ".VV.bin";
  XMLFileWriter xml(output);
 
  LatticeFermion chi,psi;

  /***************************************/
  /* Collect some useful 5d correlators  */
  /***************************************/
  LatticeComplex PA = zero;
  LatticeComplex PP = zero;
  LatticeComplex PAconsv = zero;
  LatticeComplex PJ5q    = zero;

  multi1d<int> iterations(Nc*Ns);
  multi1d<Real> residuals(Nc*Ns);

  // Optionally do the vacuum polarisation insertions
  multi1d<LatticePropagator> sol_5;
  if ( do_PImunu ) {
    sol_5.resize(Ls);
    for(int s=0;s<Ls;s++) sol_5[s]=zero;
  }
  
  for(int color_source = 0; color_source < Nc; ++color_source) {
    for(int spin_source = 0; spin_source < Ns; ++spin_source) {
	
      int idx = spin_source*3+color_source;

      // Extract a fermion source
      PropToFerm(src, chi, color_source, spin_source);

      /**************************************************************
       * Zero source implies zero solution 
       ***************************************************************
       */
      Real nrm = sqrt(norm2(chi));
      if (toFloat(nrm) == 0.0) {

	psi = zero;
	iterations[idx] = 0;
        FermToProp(psi, sol, color_source, spin_source);

      } else {

	multi1d<LatticeFermion> chi_5(Ls), psi_5(Ls);


	/*******************************************************************
	 * Do the CG :
	 *
	 * Normalize the source in case it is really huge or small -
	 * a trick to avoid overflows or underflows.
	 *
	 * Splat the chiralities of the source to the walls
	 *******************************************************************/

	Real fact = 1.0/nrm;
	chi *= fact;
	chi_5[0]   = chiralProjectPlus(chi);
	for(int s=1;s<Ls-1;s++) chi_5[s] = zero;
	chi_5[Ls-1]= chiralProjectMinus(chi);
	
	for(int s=0;s<Ls;s++) psi_5[s]=zero;

	// Call the the CG
	int iter = 
	  dwf_restarted_invert<Float_f,Float_d>(psi_5,chi_5,U,
						output_stem,
						Ls,mass,M5,
						Ncg,residual,max_iter);
	iterations[idx] = iter;

        fact = Real(1) / fact;

        for(int s=0;s<Ls;s++) psi_5[s] *= fact;

	/*******************************************************************
	 * Accumulate important 5d matrix elements
	 *******************************************************************
	 */
	      /****************************/  
	      /* Get the midpoint beastie */
	      /****************************/  

	psi = chiralProjectPlus(psi_5[Ls/2-1]);
	psi += chiralProjectMinus(psi_5[Ls/2]);
	PJ5q+= localInnerProduct(psi,psi);
	
	      /****************************/  
	      /* Get the conserved axial  */
	      /****************************/  
	LatticeComplex C;
	LatticeFermion    p5d;
	LatticeFermion us_p5d;
	int mu=Nd-1;
	int g5=Ns*Ns-1;
	int gt=8;

	for(int s=0;s<Ls;s++){

	  p5d = psi_5[Ls-1-s] ;
	  us_p5d = U[mu]*shift(psi_5[s],FORWARD,mu);

	  C = 0.5*localInnerProduct(Gamma(g5)*p5d,Gamma(gt)*us_p5d);
	  C-= 0.5*localInnerProduct(Gamma(g5)*p5d,us_p5d);

	  p5d = psi_5[s];
	  us_p5d = U[mu]*shift(psi_5[Ls-1-s],FORWARD,mu);

	  C+= 0.5*localInnerProduct(Gamma(g5)*us_p5d,Gamma(gt)*p5d);
	  C+= 0.5*localInnerProduct(Gamma(g5)*us_p5d,p5d);

	  if (s < Ls/2) PAconsv -= C;
	  else          PAconsv += C;

	}


	/****************************/  
	/* Get the full VV AA       */
	/****************************/  

	if ( do_PImunu ) {
	  for(int s=0;s<Ls;s++){
	    FermToProp(psi_5[s], sol_5[s], color_source, spin_source);
	  }
	}

	/********************************************************
	 * Merge the chiralities from the walls to reconstruct 4d prop
	 ********************************************************/
	psi = chiralProjectMinus(psi_5[0]);
	psi += chiralProjectPlus(psi_5[Ls-1]);

        FermToProp(psi, sol, color_source, spin_source);

	/*******************************************************************
	 * Accumulate important 4d matrix elements
	 *******************************************************************
	 */
	PA += localInnerProduct(psi,Gamma(1<<Nd-1)*psi);
	PP += localInnerProduct(psi,psi);

      }
    } /* end loop over spin_source */
  } /* end loop over color_source */

  
#ifdef BFM_CHROMA_BINDING
  /************************************************************************/
  /* Contract PImunu for V and A                                          */
  /************************************************************************/
  if ( do_PImunu ) {
   BinaryFileWriter WR(outputbin);

   // 3d FT
   SftMom Phases(2,false,Nd-1);
   int length = Phases.numSubsets();
   int nmom   = Phases.numMom();
   

    for(int mu=0;mu<Nd;mu++){ // Sink direction (of conserved current)
    for(int nu=0;nu<Nd;nu++){ // source direction (of local current)

      // Collect the P channel
      LatticeComplex PAc = zero;
      LatticeComplex CP;
      // AA
      LatticeComplex AAc = zero;
      LatticeComplex CA;
      // VV
      LatticeComplex VVc = zero;
      LatticeComplex CV;

      LatticePropagator p5d;
      LatticePropagator us_p5d;

      SpinMatrix G5 =1.0; G5  = G5*Gamma(15);
      SpinMatrix Gmu=1.0; Gmu = Gmu*Gamma(1<<mu);
      SpinMatrix Gnu=1.0; Gnu = Gnu*Gamma(1<<nu);

      for(int s=0;s<Ls;s++){
 
        p5d = sol_5[Ls-1-s] ;
        us_p5d = U[mu]*shift(sol_5[s],FORWARD,mu);

        CV = 0.5*trace(Gnu*G5*adj(p5d)*G5*Gmu*us_p5d);
        CV-= 0.5*trace(Gnu*G5*adj(p5d)*G5*us_p5d);

        CA = 0.5*trace(Gnu*adj(p5d)*G5*Gmu*us_p5d);
        CA-= 0.5*trace(Gnu*adj(p5d)*G5*us_p5d);

        CP = 0.5*trace(adj(p5d)*G5*Gmu*us_p5d);
        CP-= 0.5*trace(adj(p5d)*G5*us_p5d);

        p5d = sol_5[s];
        us_p5d = U[mu]*shift(sol_5[Ls-1-s],FORWARD,mu);

        CV+= 0.5*trace(Gnu*G5*adj(us_p5d)*G5*Gmu*p5d);
        CV+= 0.5*trace(Gnu*G5*adj(us_p5d)*G5*p5d);
        CA+= 0.5*trace(Gnu*adj(us_p5d)*G5*Gmu*p5d);
        CA+= 0.5*trace(Gnu*adj(us_p5d)*G5*p5d);
        CP+= 0.5*trace(adj(us_p5d)*G5*Gmu*p5d);
        CP+= 0.5*trace(adj(us_p5d)*G5*p5d);

        VVc += CV;
        if (s < Ls/2) {
	  PAc -= CP;
	  AAc -= CA;
	} else {
	  PAc += CP;
	  AAc += CA;
	}
	
      }// s Loop

      ////////////////////////////////////
      // Do the fourier transform
      ////////////////////////////////////
      multi2d<DComplex> corr(nmom,length);

      corr = Phases.sft(PAc);
      write(WR,corr);
      
      corr = Phases.sft(AAc);
      write(WR,corr);

      corr = Phases.sft(VVc);
      write(WR,corr);

    }
    }//munu

    multi1d<multi1d<int> > momlist(nmom);
    for(int m=0;m<nmom;m++) momlist[m] = Phases.numToMom(m);
    write(WR,momlist);

  }
#endif


  /************************************************************************/
  /* Spatially sum the correlation functions                              */
  /************************************************************************/
  Set &tslice = GetTimeslice();
  int length = tslice.numSubsets();
  multi1d<DComplex> corr(length);
  multi1d<Real>    rcorr(length);

  /***********************************************************************/
  /* Write the XM-HELL                                                   */
  /***********************************************************************/
  push(xml,"DWF_prop");
  write(xml,"iterations",iterations);
  write(xml,"residuals",residuals);
    push(xml,"DWF_observables");

      corr = sumMulti(PP,tslice);  
      for(int t=0;t<length;t++)rcorr[t]=real(corr[t]);
      write(xml,"PP",rcorr);

      corr = sumMulti(PA,tslice);  
      for(int t=0;t<length;t++)rcorr[t]=real(corr[t]);
      write(xml,"PA",rcorr);

      corr = sumMulti(PAconsv,tslice);  
      for(int t=0;t<length;t++)rcorr[t]=real(corr[t]);
      write(xml,"PAconsv",rcorr);

      corr = sumMulti(PJ5q,tslice);  
      for(int t=0;t<length;t++)rcorr[t]=real(corr[t]);
      write(xml,"PJ5q",rcorr);

    pop(xml);     
  pop(xml);      
  return 0; 
}


// Internal interfaces
template <class Float_f,class Float_d>
int dwf_restarted_invert(multi1d<LatticeFermion> &sol, 
			 multi1d<LatticeFermion> &src,
			 multi1d<LatticeColorMatrix> &U,
			 std::string output_stem,
			 int Ls,
			 Real mass,
			 Real M5,
			 int Ncg,
			 Real residual[],
			 int max_iter[])
{
  multi1d<int> iter(Ncg);

  multi1d<LatticeFermion> sol_f(Ls);
  std::string stem_f = output_stem + ".prop.f";
  std::string stem_d = output_stem + ".prop.d";

  for(int s=0;s<Ls;s++) sol_f[s] = zero;
  for(int cg=0;cg<Ncg-1;cg++){
    iter[cg] = 
      dwf_invert<Float_f>(sol_f,src,U,stem_f,
			  Ls,mass,M5,
			  residual[cg],max_iter[cg]);
  }
  int iter_all=0;
  for(int s=0;s<Ls;s++) sol[s] = sol_f[s];

  {
    int cg = Ncg-1;
    iter[cg] = dwf_invert<Float_d>(sol,src,U,stem_d,
				   Ls,mass,M5,
				   residual[cg],max_iter[cg]);
    iter_all+=iter[cg];
  }

  for(int cg=0;cg<Ncg;cg++){
    QDPIO::cout <<iter[cg]<<" pass " << cg<< " iterations"<<endl;
  }
  //  QDPIO::cout <<"Normalised difference is "<<sqrt(norm2(sol - sol_f)/norm2(sol))<<endl;
  return iter_all;
}

typedef LatticeFermion T;
typedef multi1d<LatticeFermion> T5;
typedef multi1d<LatticeColorMatrix> G;

// Internal interfaces
template <class Float>
int dwf_invert(multi1d<LatticeFermion> &sol, 
	       multi1d<LatticeFermion> &src,
	       multi1d<LatticeColorMatrix> &U,
	       std::string output_stem,
	       int Ls,
	       Real mass,
	       Real M5, Real residual, int max_iter)
{
  // Set up BAGEL object
  int lx = QDP::Layout::subgridLattSize()[0];
  int ly = QDP::Layout::subgridLattSize()[1];
  int lz = QDP::Layout::subgridLattSize()[2];
  int lt = QDP::Layout::subgridLattSize()[3];

  multi1d<int> procs = QDP::Layout::logicalSize();

  bfmarg dwfa;

  //Physics parameters
  dwfa.Ls           = Ls;
  dwfa.M5           = toDouble(M5);
  dwfa.mass         = toDouble(mass);
  dwfa.precon_5d    = 1;
  dwfa.reproduce    = 0;

  dwfa.max_iter     = max_iter;
  dwfa.residual     = toDouble(residual);


  //Geometry
  dwfa.node_latt[0] = lx;
  dwfa.node_latt[1] = ly;
  dwfa.node_latt[2] = lz;
  dwfa.node_latt[3] = lt;
  
  multi1d<int> ncoor = Layout::nodeCoord();
  dwfa.ncoor[0] = ncoor[0];
  dwfa.ncoor[1] = ncoor[1];
  dwfa.ncoor[2] = ncoor[2];
  dwfa.ncoor[3] = ncoor[3];


  for(int mu=0;mu<4;mu++){
    if (procs[mu]>1) dwfa.local_comm[mu] = 0;
    else             dwfa.local_comm[mu] = 1;
  }

  QDPIO::cout << "Initialising BAGEL-2 solver "<<endl;

  bfm_internal<Float> dwf; dwf.init(dwfa);

  dwf.importGauge(U);

  Fermion_t psi_h[2], chi_h[2];

  for(int cb=0;cb<2;cb++){
    psi_h[cb] = dwf.allocFermion();
    chi_h[cb] = dwf.allocFermion();
    dwf.importFermion(src,psi_h[cb],cb);
    dwf.importFermion(sol,chi_h[cb],cb);
  }

  int iter = dwf.CGNE(chi_h,psi_h);

  for(int cb=0;cb<2;cb++){
    dwf.exportFermion(sol,chi_h[cb],cb);
    dwf.freeFermion(psi_h[cb]);
    dwf.freeFermion(chi_h[cb]);
  }
  dwf.end();

#ifdef BFM_CHROMA_BINDING
  /*******************************************************/
  /* verify the solution                                 */
  /*******************************************************/

  /*
   * Compute the residual according to CHROMA
   */
  multi1d<LatticeFermion> regress(Ls);
   multi1d<int> bcs(Nd);
   bcs[0] = bcs[1] = bcs[2] = bcs[3] = 1;

   Handle< FermBC<T,G,G> > fbc(new SimpleFermBC< T, G, G >(bcs));
   Handle<CreateFermState<T,G,G> > cfs( new CreateSimpleFermState<T,G,G>(fbc));

   UnprecDWFermActArray  S_f(cfs, M5, mass, Ls);
   Handle< FermState<T,G,G> > fs( S_f.createState(U) );

   Handle< UnprecLinearOperatorArray<T,G,G> > M(S_f.unprecLinOp(fs,mass));
  
   // Check the result
   PlusMinus pm = PLUS;

   (*M)(regress,sol,pm);

   regress = regress - src;
   QDPIO::cout << "QDP regression check :  |M sol - src| = " << norm2(regress) << endl;
#endif

   return iter;

}


template <class Float>
int bfmbase<Float>::CG4d(LatticePropagator &sol, 
			  LatticePropagator &src,
			  multi1d<LatticeColorMatrix> &U,
			  std::string output_stem)
{


  Fermion_t chi_h[2],psi_h[2];

  this->importGauge(U);

  for(int cb=0;cb<2;cb++){
    chi_h[cb] = this->allocFermion();
    psi_h[cb] = this->allocFermion();
  }

  for(int color_source = 0; color_source < Nc; ++color_source) {
    for(int spin_source = 0; spin_source < Ns; ++spin_source) {
	
      LatticeFermion psi = zero;
      LatticeFermion chi;

      // Extract a fermion source
      PropToFerm(src, chi, color_source, spin_source);

      /*
       * Normalize the source in case it is really huge or small -
       * a trick to avoid overflows or underflows
       */
      Real nrm = sqrt(norm2(chi));
      if (toFloat(nrm) == 0.0) {
	psi = zero;
      } else {
	Real fact = 1.0/nrm;
	chi *= fact;

	for(int cb=0;cb<2;cb++){
	  this->importFermion(chi,chi_h[cb],cb);
	  this->importFermion(psi,psi_h[cb],cb);
	}
	int iter = this->CGNE(psi_h, chi_h);
	
	for(int cb=0;cb<2;cb++){
	  this->exportFermion(psi,psi_h[cb],cb);
	}	
        // Unnormalize the source following the inverse of the normalization above
        fact = Real(1) / fact;
        psi *= fact;

        /*
         * Move the solution to the appropriate components
         * of quark propagator.
         */
        FermToProp(psi, sol, color_source, spin_source);
      }

    } /* end loop over spin_source */
  } /* end loop over color_source */

  for(int cb=0;cb<2;cb++){
    this->freeFermion(psi_h[cb]);
    this->freeFermion(chi_h[cb]);
  }
  return 0;
}

/*******************************************************************
 * This is the QDP glue layer
 *******************************************************************
 */

#include <qdp.h>


// This is the base Bagel 2 import. Use QDP to do the shifting, conjugation why not....
template <class Float>
void      bfmbase<Float>::importGauge(multi1d<LatticeColorMatrix> &u)
{
  if ( u.size() != Nd ) exit(-1);
  for ( int mu=0;mu<Nd;mu++){

    QDPdouble * U_p;

    int dir = 2*mu+1;
    U_p = (QDPdouble *)&u[mu].elem(0).elem();
    importGauge(U_p,dir);

    dir = 2*mu;
    LatticeColorMatrix udag;
    udag = adj(shift(u[mu],BACKWARD,mu));
    U_p = (QDPdouble *)&udag.elem(0).elem();
    importGauge(U_p,dir);

  }
}

template <class Float>
void      bfmbase<Float>::importFermion(LatticeFermion & psi, Fermion_t handle, int cb,int s)
{
// Need to worry about base parity 
  QDPdouble *psi_p = (QDPdouble *) &psi.elem(0).elem(0).elem(0).real();
  impexFermion(psi_p,handle,1,cb,s);
}
template <class Float>
void      bfmbase<Float>::exportFermion  (LatticeFermion &psi, Fermion_t handle, int cb,int s)
{
// Need to worry about base parity
  QDPdouble *psi_p = (QDPdouble *) &psi.elem(0).elem(0).elem(0).real();
  impexFermion(psi_p,handle,0,cb,s);
}

template <class Float>
void      bfmbase<Float>::importFermion(multi1d<LatticeFermion> &psi, Fermion_t handle, int cb)
{
// Need to worry about base parity 
  if ( psi.size() != Ls ) exit(-1);
  for(int s=0;s<Ls;s++){
    QDPdouble *psi_p = (QDPdouble *) &(psi[s].elem(0).elem(0).elem(0).real());
    impexFermion(psi_p,handle,1,cb,s);
  }
}
template <class Float>
void      bfmbase<Float>::exportFermion  (multi1d<LatticeFermion> &psi, Fermion_t handle, int cb)
{
// Need to worry about base parity
//  cout << "Export 5d fermion "<<Ls<<endl;
  if ( psi.size() != Ls ) exit(-1);
  LatticeFermion tmp;
  for(int s=0;s<Ls;s++){
    QDPdouble *psi_p = (QDPdouble *) &(psi[s].elem(0).elem(0).elem(0).real());
    impexFermion(psi_p,handle,0,cb,s);
  }
}


#endif
#endif
