#include <stdlib.h>
#include <stdio.h>

#include "bfm.h"
#include "bfmbagel1.h"
#include "bfmcommfakebagel1.h"

template<class Float>
void commfakebagel1<Float>::comm_init (void)
{
  if (this->isBoss() ) printf("Initialising FAKE communications\n");
}

template<class Float>
void commfakebagel1<Float>::comm_end (void)
{

}

template<class Float>
void commfakebagel1<Float>::comm_start(int result_cb, Fermion_t psi,int dag)
{
  int idx,atom;
  int base,offset;
  char *Base;
  int stride;

  char *ptr = (char *) this->two_spinor;

  for ( int mu=0;mu<ND;mu++ ) { 

    if ( !this->local_comm[mu] ) { 

	// Normal receives
	// Just copy one set of buffers to the other
	for (int  mp=0;mp<2;mp++) {	  

	  // Do the fake "comms"
	  Float *rbuf = this->recv_bufs[result_cb][1-mp][mu];
	  Float *sbuf = this->send_bufs[result_cb][mp][mu];
	  for(idx=0;idx<this->simd_nbound[mu];idx++){
	    int offset = idx *this->PAD_HALF_SPINOR_SIZE;
	    memcpy((char *)&rbuf[offset],
		   (char *)&sbuf[offset],
		   HALF_SPINOR_SIZE*sizeof(Float)); 
	  }
	  
	}
    }
  }
}

template<class Float>
void commfakebagel1<Float>::comm_complete(int result_cb, Fermion_t psi )
{
  for ( int mu=0;mu<ND;mu++ ) { 
    if ( !this->local_comm[mu] ) { 
      for (int mp=0;mp<2;mp++) {

	double *dp =(double *) this->recv_bufs[result_cb][mp][mu];
	printf("%lx -> %lx %le\n",this->two_spinor[result_cb],this->recv_bufs[result_cb][mp][mu],
	       *dp);
	               
	if ( sizeof(Float) == sizeof(double) ) {
	  vmx_face_scatter((double *)this->two_spinor[result_cb],
			   (double *)this->recv_bufs[result_cb][mp][mu],
			   this->face_table[result_cb][1-mp][mu],
			   this->simd_nbound[mu]);
	} else {
	  vmx_face_scatter_single((float *)this->two_spinor[result_cb],
				  (float *)this->recv_bufs[result_cb][mp][mu],
				  this->face_table[result_cb][1-mp][mu],
				  this->simd_nbound[mu]);
	}
	
      }
    }
  }
  return;
}

template<class Float>
void commfakebagel1<Float>::comm(int result_cb, Fermion_t psi, int dag)
{
  this->comm_start(result_cb,psi,dag);
  this->comm_complete(result_cb,psi);
}
template class commfakebagel1<float>;
template class commfakebagel1<double>;
