
#include <stdlib.h>
#include <stdio.h>

#include "bfm.h"
#include "bfmbagel1.h"
#include "bfmcommqmpbagel1.h"

template<class Float>
void commQMPbagel1<Float>::comm_init (void)
{
  num_op = 0;

  if ( this->isBoss() ) { 
    printf("commQMPbagel1::comm_init()\n");
  }
  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++ ) {
      if( !this->local_comm[mu] ) {
	for(int cb=0;cb<Ncb;cb++){

	  int nblock= this->simd_nbound[mu];
	  int block =     HALF_SPINOR_SIZE *sizeof(Float);
	  int stride= this->PAD_HALF_SPINOR_SIZE *sizeof(Float);
	  
	  int delta;
	  if ( pm == Plus ) delta = 1;
	  else delta = -1;

	  int idx=2*mu+pm; 
	  this->recv_ops_msgmem_t[cb][idx] = QMP_declare_strided_msgmem((void *)this->recv_bufs[cb][pm][mu],
									block,
									nblock,
									stride);
	  this->recv_handles[cb][num_op]  = QMP_declare_receive_relative(recv_ops_msgmem_t[cb][idx], mu, delta, 0);

	  this->send_ops_msgmem_t[cb][idx] = QMP_declare_strided_msgmem((void *)this->send_bufs[cb][pm][mu],
								    block,
								    nblock,
								    stride);
	  send_handles[cb][num_op]  = QMP_declare_send_relative(send_ops_msgmem_t[cb][idx], mu, delta, 0);

	}
	num_op ++;
      }
    }
  } 
  if ( this->isBoss() ) {
    printf("comm_init: %d send ops\n",num_op);
  }
  if( num_op > 0 ) { 
    for(int cb=0;cb<Ncb;cb++){
      send_multi_handle[cb] = QMP_declare_multiple(send_handles[cb], num_op);
      recv_multi_handle[cb] = QMP_declare_multiple(recv_handles[cb], num_op);
    }
  }
}

template<class Float>
void commQMPbagel1<Float>::comm_end (void)
{
  if ( num_op>0 ) {
    for(int cb=0;cb<2;cb++){
      QMP_free_msghandle(send_multi_handle[cb]);
      QMP_free_msghandle(recv_multi_handle[cb]);
    }
  }
  for(int mu=0; mu < ND; mu++) { 
    if (!this->local_comm[mu]) { 
      for(int cb=0;cb<2;cb++){
	for(int pm=0;pm<2;pm++){
	  QMP_free_msgmem(send_ops_msgmem_t[cb][2*mu+pm]);
	  QMP_free_msgmem(recv_ops_msgmem_t[cb][2*mu+pm]);
	}
      }
    }
  }
}

template<class Float>
void commQMPbagel1<Float>::comm_start(int result_cb, Fermion_t psi,int dag)
{
  if ( num_op > 0 ) {
    QMP_start(send_multi_handle[result_cb]);
    QMP_start(recv_multi_handle[result_cb]);
  }
}

template<class Float>
void commQMPbagel1<Float>::comm_complete(int result_cb, Fermion_t psi )
{
  if ( num_op > 0 ) { 

    if( QMP_wait(recv_multi_handle[result_cb]) != QMP_SUCCESS ) { 
      if ( isBoss() ) printf("QMP_wait returned unsuccessful value: \n");
      exit(-1);
    }
    
    if( QMP_wait(send_multi_handle[result_cb]) != QMP_SUCCESS ) { 
      if ( isBoss() ) printf("QMP_wait returned unsuccessful value: \n");
      exit(-1);
    }

  }
  
  for ( int mu=0;mu<ND;mu++ ) { 
    if ( !this->local_comm[mu] ) { 
      for (int mp=0;mp<2;mp++) {

	if ( sizeof(Float) == sizeof(double) ) {
	  vmx_face_scatter((double *)this->two_spinor[result_cb],
			   (double *)this->recv_bufs[result_cb][mp][mu],
			   this->face_table[result_cb][1-mp][mu],
			   this->simd_nbound[mu]);
	} else {
	  vmx_face_scatter_single((float *)this->two_spinor[result_cb],
				  (float *)this->recv_bufs[result_cb][mp][mu],
				  this->face_table[result_cb][1-mp][mu],
				  this->simd_nbound[mu]);
	}
	
      }
    }
  }
  return;
}

template<class Float>
void commQMPbagel1<Float>::comm(int result_cb, Fermion_t psi, int dag)
{
  this->comm_start(result_cb,psi,dag);
  this->comm_complete(result_cb,psi);
}

template <class Float> 
void commQMPbagel1<Float>::comm_gsum(double &val) 
{
  QMP_sum_double_array(&val,1);
}
template <class Float> 
bool commQMPbagel1<Float>::isBoss(void)
{
  return QMP_is_primary_node();
}

template class commQMPbagel1<float>;
template class commQMPbagel1<double>;
