#include <chroma.h>
#include <actions/ferm/invert/syssolver_linop_cg_array.h>
#include <bfm.h>
#include <bfm_qdp.h>

typedef bfm_dp bfm_t;

using namespace Chroma;

typedef LatticeFermion T;
typedef multi1d<LatticeFermion> T5;
typedef multi1d<LatticeColorMatrix> U;

//#define Printf if ( QMP_is_primary_node() ) printf
#define Printf printf

int main (int argc,char **argv )
{
  Chroma::initialize(&argc,&argv);


  /********************************************************
   * Command line parsing
   ********************************************************
   */
  if ( argc != 6 ) { 
   Printf("Usage: %s lx ly lz lt Ls\n All must be even\n",argv[0]);
   Printf("argc is %d\n",argc);
   for ( int i=0;i<argc;i++){
      Printf("%d %s\n",i,argv[i]);
   }
      exit(-1);

  }

  /********************************************************
   * Setup QDP
   ********************************************************
   */
  multi1d<int> nrow(Nd);
  nrow[0] = atoi(argv[1]);
  nrow[1] = atoi(argv[2]);
  nrow[2] = atoi(argv[3]);
  nrow[3] = atoi(argv[4]);

  Layout::setLattSize(nrow);
  Layout::create();

  int Ls = atoi(argv[5]);

  /********************************************************
   * Gaussian gauge field
   ********************************************************
   */

  multi1d<LatticeColorMatrix>  u(Nd); HotSt(u);
  multi1d<LatticeColorMatrixF> uf(Nd);
  for(int m=0; m < Nd; ++m){
    uf[m] = u [m];
    u [m] = uf[m];
  }

  /********************************************************
   * Gaussian source and result vectors
   ********************************************************
   */
  multi1d<LatticeFermion> psi(Ls);
  multi1d<LatticeFermion> chi(Ls);
  multi1d<LatticeFermion> chi_qdp(Ls);

  for(int s=0;s<Ls;s++) psi[s]=zero;
  for(int s=0;s<Ls;s++) gaussian(chi[s]);
  for(int s=0;s<Ls;s++) gaussian(chi_qdp[s]);
  gaussian(psi[0]);
  gaussian(psi[Ls-1]);

  /********************************************************
   * Import gauge field to BAGEL
   ********************************************************
   */
  LatticeFermion noise; gaussian(noise);


  Real M5 = 1.8;
  Real mass = 0.00;
  Real mq = mass;

#define Ncg 3
  Real residuals[Ncg] ; 
  int  max_iter[Ncg];
  residuals[0] = 1.0e-6;
  residuals[1] = 1.0e-7;
  residuals[2] = 1.0e-8;
  //  residuals[3] = 1.0e-8;

  for(int i=0;i<Ncg;i++) max_iter[i]  = 1000;

  dwf_restarted_invert<float,double>(chi, 
				     psi,
				     u,
				     "/host2/pab/pooh.xml",
				     Ls,
				     mass,
				     M5,Ncg,residuals,max_iter);

 
#define CHECKIT
#ifdef  CHECKIT 
  /********************************************************
   * QDP Linop
   ********************************************************
   */
   multi1d<int> bcs(Nd);
   bcs[0] = bcs[1] = bcs[2] = bcs[3] = 1;

   Handle< FermBC<T,U,U> > fbc(new SimpleFermBC< T, U, U >(bcs));
   Handle<CreateFermState<T,U,U> > cfs( new CreateSimpleFermState<T,U,U>(fbc));
   EvenOddPrecDWFermActArray  S_f(cfs, M5, mq, Ls);
   Handle< FermState<T,U,U> > fs( S_f.createState(u) );

   Handle< EvenOddPrecLinearOperatorArray<T,U,U> > M(S_f.precLinOp(fs,mq));
   Handle< LinearOperatorArray<T> > HM(S_f.precLinOp(fs,mq));

  // Check the result
  SysSolverCGParams invParam;
  invParam.RsdCG        = residuals[1];
  invParam.RsdCGRestart = residuals[1];
  invParam.MaxCG        = max_iter[1];
  invParam.MaxCGRestart = max_iter[1];

  Handle<LinOpSystemSolverArray<T> > Hsolver = 
    new LinOpSysSolverCGArray<T>(HM,invParam);
  Handle<PrecFermAct5DQprop<T,U,U> > solver =
    new PrecFermAct5DQprop<T,U,U>(M,Hsolver);
  SystemSolverResults_t  stats =  (*solver)(chi_qdp,psi);
  int n_count = stats.n_count;

  Double n2 = 0.0;
  for(int s=0;s<Ls;s++) { 
    Double sn2 = norm2(chi[s]-chi_qdp[s]);
    n2+=sn2;
#ifdef DEBUG
    QDPIO::cout << "|| Bagel - QDP || ["<<s<<"] = "<< sn2 << endl;
#endif
  }
  QDPIO::cout << "|| Bagel - QDP || = "<< n2 << endl;
  
  n2 = 0.0;
  for(int s=0;s<Ls;s++) { 
    Double sn2 = norm2(chi[s]);
    n2+=sn2;
#ifdef DEBUG
    QDPIO::cout << "|| Bagel || ["<<s<<"] = "<< sn2 << endl;
#endif
  }
  QDPIO::cout << "|| Bagel || = "<< n2 << endl;
  
  n2 = 0.0;
  for(int s=0;s<Ls;s++) {
    Double sn2 = norm2(chi_qdp[s]);
    n2+=sn2;
#ifdef DEBUG
    QDPIO::cout << "|| QDP || ["<<s<<"] = "<< sn2 << endl;
#endif
  }
  QDPIO::cout << "|| QDP || = "<< n2 << endl;
#endif

  Printf("Done\n"); 

}








